// Ajax headers
#include <Wire.h>
#include <Arduino.h>
#include <WiFi.h>
#include <WiFiClient.h>
#include <WebServer.h>
#include "index.h"
// LoRa headers
#include <SPI.h>
#include <LoRa.h>


// Pin selections for LoRa SPI -- need to be changed for our board
#define ss 5
#define rst -1
#define dio0 20
// Pin selections for actuators
#define LED1 0
#define LED2 1
#define LED3 1

WebServer server(80);
const char* ssid = "ND-guest";
const char* password = "";

// Variable Declarations
bool LED1status = LOW;
bool LED2status = LOW;
bool LED3status = LOW;
bool messageComp = 0;
bool confirmed = false;

int message2send = 0;
int lastMessage = 0;

byte deviceID = 0x02;
byte baseID = 0x01;
byte act1val = 1;
byte act2val = 0; // set to zero for now since we don't have capability to change other things
byte act3val = 0;
byte act4val = 0;
byte ack = 0xAA;
byte nack = 0x00;
byte req = 0x99;

// uint8_t part1con;

int sensorID;
int checkBattery;
int analogSensor;
int lightSensor;
int i2c2Val;
int i2c3Val;
int i2c4Val;
int analSense;
int vemlRead;
int humRead;

String testServer = "a";
String testArray;
String statusUpdate;


// Functions for server

void handleRoot() {
 String s = MAIN_page; //Read HTML contents
 server.send(200, "text/html", s); //Send web page
}

void handleADC() {
 server.send(200, "text/plane", testArray); //Send ADC value only to client ajax request
 testArray[0] = 0x69; //arbitrary value that is not 0x01 like the basestation value
 // delay(1000);
}


void act1Control() {
  String state = "ON";
  String act_state = server.arg("state");
  deviceID = 0x02;
  if(act_state == "1"){
    act1val = 0xFF;
    state = "ON";
    Serial.print("on ");
  }
  else {
    act1val = 0x00;
    state = "OFF";
    Serial.print("off ");
  }
  server.send(200, "text/plane", state);
  
  confirmed = false;
  while (!confirmed) {
    packetWrite(nack);
    //delay(500)
    Serial.println("NACK message sent");
  }
  
  Serial.println("Packet Sent");
}

void updateRequest() {
  String deviceNumber = server.arg("device");
  if(deviceNumber == "3"){ Serial.println("Requesting from Device 3");
    deviceID = 0x03;
  }
  else if (deviceNumber == "2"){ Serial.println("Requesting from Device 2");
    deviceID = 0x02;
  }
  confirmed = false;
  while(!confirmed){
    packetWrite(req);
    Serial.println("Status Request Sent");
  }
  
  server.send(200, "text/plane", statusUpdate);
  // delay(1000); 
}


void sprinkSel() {
  String selectedSprink = server.arg("state");
  Serial.print("sprink: ");
  Serial.println(selectedSprink);
  if(selectedSprink == "1"){
    act3val = 0xFF;
  }
  else{
    act3val = 0x00;
  }

  confirmed = false;
  while(!confirmed){
  packetWrite(nack);
  Serial.println("NACK message sent");
  }
  
}

void deviceSel() {
  String selectedDevice = server.arg("state");
  Serial.print("device: ");
  Serial.println(selectedDevice);
  if(selectedDevice == "3"){
    deviceID = 0x03;
  }
  else if(selectedDevice == "2"){
    deviceID = 0x02;
  }
  
}

void heatSel() {
  String selectedHeat = server.arg("state");
  Serial.print("heat: ");
  Serial.println(selectedHeat);
  if(selectedHeat == "1"){
      act1val = 0xFF;
      Serial.println("act1 on");
    }
    else{
      act1val = 0x00;
      Serial.println("act1 off");
    }
    
}

void shadeSel() {
  String selectedShade = server.arg("state");
  Serial.print("shade: ");
  Serial.println(selectedShade);
  if(selectedShade == "1"){
    act2val = 0xFF;
  }
  else{
    act2val = 0x00;
  }
  
}


extern "C" void phy_bbpll_en_usb(bool en);

void setup() {
  // Serial Setup:
  Serial.begin(115200); // Begin serial connection
  Serial.println("starting");
  // delay(5000);

 // WiFi Setup:
 phy_bbpll_en_usb(true);
 
 WiFi.mode(WIFI_STA); // connects to existing network
 phy_bbpll_en_usb(true);
 Serial.println("This boy's a wiFi station");
  WiFi.begin(ssid, password);
  Serial.println("Connecting to ");
  Serial.print(ssid);
  while(WiFi.waitForConnectResult() != WL_CONNECTED){      
      Serial.print("."); // wait until wifi connection established
    }
  Serial.println("");
  Serial.print("Connected to ");
  Serial.println(ssid);
  Serial.print("IP address: ");
  Serial.println(WiFi.localIP());
  phy_bbpll_en_usb(true);

  // Server Setup:
  server.on("/", handleRoot);      // Display page
  server.on("/readADC", handleADC); //check for updated packet values
  server.on("/dev_sel", deviceSel); // use this to select which device actuation is being performed on
  server.on("/heat_sel", heatSel); // pass the value of the heatlamp actuator
  server.on("/shade_sel", shadeSel); //pass value of shade actuator
  server.on("/sprink_sel", sprinkSel); //pass value of sprinkler actuator
  server.on("/request", updateRequest); // call this function when you want a live update and push the update button
  
   server.begin(); //Start server
   Serial.println("HTTP server started");

  // Pin Mode Setup:
  pinMode(LED1,OUTPUT);
  pinMode(LED2,OUTPUT);
  pinMode(LED3,OUTPUT);
  digitalWrite(LED1,HIGH); // need to change from digital to analog write (everything)
  digitalWrite(LED2,HIGH);
  digitalWrite(LED3,HIGH);

  // LoRa Setup:
  LoRa.setPins(ss,rst,dio0);
  while (!LoRa.begin(915E6)) { // waits until LoRa is set up
    Serial.println(".");
    delay(500);
  }
  LoRa.setSyncWord(0x01); // receive only from paired sender
  Serial.println("LoRa Worked");
}

void loop() {
  
  // put your main code here, to run repeatedly:
  server.handleClient(); //packetWrite happens here when server UI is utilized
  delay(1);
  readPacket(); 
  // ****
//  int button = analogRead(0);
//  if (button < 100) {
//    delay(50);
//    Serial.print(button);
//    confirmed = false;
//    while (!confirmed) {
//      packetWrite(0x00);  
//    }
//  }
}

// prepare basestation to field device packet
void packetWrite(byte acknowledgeYN) {

  Serial.println("packetWrite called");
  Serial.print("act1 val: ");
  Serial.println(act1val);
  Serial.print("act2 val: ");
  Serial.println(act2val);
  Serial.print("act3 val: ");
  Serial.println(act3val);
  
  // create packet
  LoRa.beginPacket();
  LoRa.write(deviceID);
  LoRa.write(act1val);
  LoRa.write(act2val);
  LoRa.write(act3val);
  LoRa.write(act4val);
  LoRa.write(acknowledgeYN);
  LoRa.endPacket();

  // delay(100);
  
  //check for confirmation
  if (acknowledgeYN != ack) {
//    if(acknowledgeYN == req){
//      for (int i = 1; i <= 10; i++){
//        checkSend();
//      }
//    }
    checkSend();
  }
}
  


void readPacket() { 
  int packetSize = LoRa.parsePacket();
  // Serial.println("entered readPacket");
  if (packetSize) { Serial.println("Package found");
    while (LoRa.available()) { 
      String message = LoRa.readString();
      Serial.println("Homies we gotsta package");
      if (baseID == message[0]) {
        if (message[11] != ack) { //needs to be 11c
          testArray = message;
          int deviceIDnum = message[1];
          deviceID = deviceIDnum;
          // testServer = deviceID;
          int batteryval = message[2];
      
          int analsens1 = message[3];
          int analsens2 = message[4];
          analSense = bitShift(analsens1, analsens2);
          
          int vemlread1 = message[5];
          int vemlread2 = message[6];
          vemlRead = bitShift(vemlread1, vemlread2);
      
          int humread1 = message[7];
          int humread2 = message[8];
          humRead = bitShift(humread1, humread2);

          int tempread1 = message[9];
          int tempread2 = message[10];
          int tempRead = bitShift(tempread1, tempread2);
       
          Serial.println(message);
          Serial.println("Base Station ID");
          Serial.println(baseID);
          Serial.println("Device ID");
          Serial.println(deviceIDnum);
          Serial.println("Battery Value");
          Serial.println(batteryval);
          Serial.println("Analog Sensor Reading");
          Serial.println(analsens1);
          Serial.println(analsens2);
          Serial.println("VEML Light Reading");
          Serial.println(vemlread1);
          Serial.println(vemlread2);
          Serial.println("Moisture Reading");
          Serial.println(humread1);
          Serial.println(humread2);
          Serial.print("Temp hiByte: ");
          Serial.println(tempread1, HEX);
          Serial.println("Temp loByte: ");
          Serial.println(tempread2, HEX);
          Serial.print("Temp Val: ");
          Serial.print(tempRead);

          for(int i = 1; i < 14; i++) {packetWrite(ack);}
          
        }
      }
    }
  }
}


// Functions for Transmission Reliability

//checkSend is used for confirming that a sent message has been received
// both base station and field device can use checkSend to ensure that 
// their message has been received

void checkSend() {

Serial.println("Entered checkSend");
 for (int i = 0; i < 15000 ; i++) { //was 8096
    int packetSize = LoRa.parsePacket();
    if (packetSize) { Serial.println("Packet Received");
      while (LoRa.available()) {
        String confirmationMessage = LoRa.readString();
        
        if (baseID == confirmationMessage[0]) {
          Serial.println("Base ID correct");
          if (confirmationMessage[11] == ack) { //needs to be 11
            Serial.println("confirmation");
            confirmed = true;
            Serial.println("Confirmed set to true! Stop sending");
            i = 15000;
          }
          else if (confirmationMessage[11] == nack){ Serial.println("request complete");
            statusUpdate = confirmationMessage;
            confirmed = true;
            i = 15000;
          }
        }
      }
    }
  }
}


int bitShift(int a, int b) {

  int shifted = ((a << 7) | b);
  return shifted;
}


/* Packet Protocol -- 15 bytes in packet -- field to basestation
 *  Address - 1 bytes
 *  AnalogIn - 3 bytes (2 AnalogIn, each 12-bit adc)
 *  I2C - 8 bytes (2 bytes per actuator, 16-bit resolution)
 */

// Example: { 0x01 0x8A932F 0x2179ABCD10006969 }

/* Packet Protocol -- basestation to field
 *  Address - 1 bytes -- intended receiver
 *  Actuator - 3 bytes (1 bytes per actuator, 8-bit desired PWM)
 */

 // Example: { 0x03 0x001111 }
