#include <SPI.h>
#include <LoRa.h>
#include <Arduino.h>
#include <SensirionI2CSht4x.h>
#include <Wire.h>
#include <SparkFun_VEML7700_Arduino_Library.h>
#include <esp32-hal-timer.h>

// Ambient Light Sensor
VEML7700 lightSensor;
// Temp / Humidity
SensirionI2CSht4x sht4x;

// Pin selections for LoRa SPI -- need to be changed for our board
#define ss 5 // 5 // 5 //16
#define rst -1 //4
#define dio0 20 //1 //20 // 26
// Pin selections for actuators
#define act1 0
#define act2 1
#define act3 1
#define act4 1

// Other pin selections
#define batteryPin 2
#define arPin 3

// for testing
int count = 0;

// incoming variables
byte deviceID = 0x03; //02
byte act1val = 0xFF;
byte act2val = 0xFF;
byte act3val;
byte act4val;
byte ack = 0xAA;
byte nack = 0x00;
byte req = 0x99;
byte firstByteAR;
byte secondByteAR;
byte firstByteLIGHT;
byte secondByteLIGHT;
byte firstByteTEMP;
byte secondByteTEMP;
byte firstByteHUM;
byte secondByteHUM;

// outgoing variables
byte baseID = 0x01; //01
byte batteryLevel;
int light;
int temperature;
int humidity;
int i2c2Val = 0xFFFF;
int i2c3Val = 0x4344;
int i2c4Val = 0x2941;

bool confirmed;
bool sht4xPresent = false;
bool luxPresent = false;

float luxRaw;
int lux;

String sensorReading;
byte acknowledge;

// timer variables
volatile int interruptCounter;
int totalInterruptCounter;

extern "C" void phy_bbpll_en_usb(bool en);

// timer initialization
static hw_timer_t *timer = NULL;
portMUX_TYPE timerMux = portMUX_INITIALIZER_UNLOCKED;
void IRAM_ATTR onTimer() {
  portENTER_CRITICAL_ISR(&timerMux);
  interruptCounter++;
  portEXIT_CRITICAL_ISR(&timerMux);
}

void setup() {
  // Serial Setup:
  phy_bbpll_en_usb(true);
  vTaskDelay(200);
  Serial.begin(115200); // Begin serial connection
  Serial.println("start");
  // Timer Setup as Interrupt
  timer = timerBegin(0, 80, true);
  timerAttachInterrupt(timer, &onTimer, true);
  timerAlarmWrite(timer, 30000000, true);
  timerAlarmEnable(timer);
  Serial.println();
  Serial.println("Got here at least");

  pinMode(act1, OUTPUT);
  pinMode(act2, OUTPUT);
  digitalWrite(act1, HIGH);
  digitalWrite(act2, HIGH);

  // Pin Mode Setup:
//  ledcSetup(0, 5000, 8); // setup led channel (channel, frequency, resolution)
//  ledcAttachPin(act1, 0); // attach pin to led channel (pin, channel)
//  ledcWrite(act1, 255); // analog write equivalent (pin, magnitude)
//  ledcSetup(1, 5000, 8);
//  ledcAttachPin(act2, 1);
//  ledcWrite(act2, 255);
//  ledcSetup(2, 5000, 8);
//  ledcAttachPin(act3, 0);
//  ledcWrite(act3, 255);
//  ledcSetup(3, 5000, 8);
//  ledcAttachPin(act4, 0);
//  ledcWrite(act4, 255);


  Serial.println("Setup Pin Mode");

  // LoRa Setup:
  //LoRa.setPins(ss,rst,dio0); // sets pins used for lora -- ss is chip select, rst is reset (may be useful for periodic reset of radio) and dio0 is used to communicate between radio and esp32
  LoRa.setPins(ss, rst, dio0);
  while (!LoRa.begin(915E6)) { // waits until LoRa is set up
    Serial.println(".");
    delay(500);
  }

  Serial.println("LoRa.setPins executed");
  Serial.println("LoRa worked");
  LoRa.setSyncWord(0x01); // receive only from paired sender
  Serial.println("LoRa set sync");
  // Light Sensor Setup:
  if (luxPresent == true) {

    Wire.begin(); // i2c initialization
    Serial.println("Wire began");

    if (lightSensor.begin() == false) { // check if VEML7700 is attached properly to the board
      Serial.println("Unable to communicate with VEML7700");
      //    while (1) {
      //      Serial.print("shit");
      //      delay(500);
      //    }
    }

    Serial.print("Light Sensor began");
    lightSensor.setIntegrationTime(VEML7700_INTEGRATION_25ms); // set integration time, modify to daylight conditions to maximize resolution without overflow
    lightSensor.setSensitivityMode(VEML7700_SENSITIVITY_x1); // set sensitivity, again may modify to maximize resolution
    Serial.println("Light sensor setup good");
    // Temp/Humidity Setup
    // Wire.begin(); // i2c initialization
  }

  delay(100);

//    // Error checking for SHT-4x:
//    uint16_t error;
//    char errorMessage[256];
//    sht4x.begin(Wire); // begin SHT-4x
//    uint32_t serialNumber;
//    error = sht4x.serialNumber(serialNumber); // if there is an error message, display, otherwise display serial number of device
//    if (error) {
//        Serial.print("Error trying to execute serialNumber(): ");
//        errorToString(error, errorMessage, 256);
//        Serial.println(errorMessage);
//    } else {
//        Serial.print("Serial Number: ");
//        Serial.println(serialNumber);
//    }
}




void loop() {
  readPacket(); // scan for received transmission

  if (interruptCounter > 0) {
    portENTER_CRITICAL(&timerMux);
    interruptCounter--;
    portEXIT_CRITICAL(&timerMux);

    totalInterruptCounter++;

    Serial.println("An interrupt has occurred. Now going to start sending process. ");
    confirmed = false;
    while (!confirmed) {
      obtainData();
      sendPacket(nack);
    }


  }

  // check for send condition
  //  int button = analogRead(0);
  //  if (button < 100) {
  //    delay(50);
  //    Serial.println("Packet Send Requested by User");
  //    confirmed = false;
  //    while (!confirmed) {
  //    obtainData();
  //    sendPacket(0x00);
  //    }
  //  }
}

void readPacket() {
  int packetSize = LoRa.parsePacket(); // check for incoming packet
  if (packetSize) { // if there is a packet:
    Serial.println("Homie my cracka is a dog (that is to say we gotsa package");
    while (LoRa.available()) {
      String receivedMessage = LoRa.readString(); // save packet as an array of bytes
      if (deviceID == receivedMessage[0]) { // first byte within the message contains the device ID. If the ID matches the field device message, then the field device takes the message, otherwise, skip
        // ****************** CHANGED THESE IF CONDITIONS TOO *****************
        if (receivedMessage[5] == nack) {
          Serial.println("nack message received, setting actuators"); // check if the message is an acknowledgement message. If acknowledgement message, then do not respond. If not an acknowledgement message, it is a data message -- save data and respond
          act1val = receivedMessage[1]; // value for actuator 1
          Serial.println(act1val);
          act2val = receivedMessage[2]; // value for actuator 2
          Serial.println(act2val);
          act3val = receivedMessage[3]; // value for actuator 3
          Serial.println(act3val);
          act4val = receivedMessage[4]; // value for actuator 4
          Serial.println(act4val);
         
                    if (act1val > 100) {
                      act1val = 1;
                    }
                    else {
                      act1val = 0;
                    }
                    if (act2val > 100) {
                      act2val = 1;
                    }
                    else {
                      act2val = 0;
                    }
          
          
                    digitalWrite(act1,act1val);
                    digitalWrite(act2,act2val);

          for (int i = 0; (i <= 15); i++) {
            sendPacket(ack); // send acknowledgement packet 15 times to ensure acknowledge is received by base station
            delay(2); // pause for 2 milliseconds for message to finish sending
          }
        }
        // ************* NEW STUFF ****************
        else if (receivedMessage[5] == req) {
          Serial.println("status request received");
          obtainData();
          for (int i = 0; (i <= 15); i++) {
            Serial.println("Status update sent");
            sendPacket(nack); // send acknowledgement packet 15 times to ensure acknowledge is received by base station
            delay(2); // pause for 2 milliseconds for message to finish sending
          }
        }

      }
    }
  }
}

void obtainData() {
  Serial.println("obtaining Data");
  checkLight();
  checkTemperatureHumidity();
  checkAR(arPin);
  checkBattery();
}

void sendPacket(byte acknowledgeYN) {
  LoRa.beginPacket();
  LoRa.write(baseID); // basestation ID -- 0
  LoRa.write(deviceID); // field device ID number -- 1
  LoRa.write(batteryLevel); // battery level reading -- 2
  LoRa.write(firstByteAR); // analog sensor high byte -- 3
  LoRa.write(secondByteAR); // analog sensor low byte -- 4
  LoRa.write(firstByteLIGHT); // light sensor high byte -- 5
  LoRa.write(secondByteLIGHT); // light sensor low byte -- 6
  LoRa.write(firstByteHUM); // humidity sensor high byte -- 7
  LoRa.write(secondByteHUM); // humidity sensor low byte -- 8
  LoRa.write(firstByteTEMP); // temperature sensor high byte -- 9
  LoRa.write(secondByteTEMP); // temperature sensor low byte -- 10
  LoRa.write(acknowledgeYN);  // acknowledgement -- 11
  LoRa.write(act1val); // actuator values -- 12
  LoRa.write(act2val);  //actuator values -- 13
  LoRa.write(act3val); //actuator values -- 14
  LoRa.endPacket();

  Serial.println("Sent!");

  if (acknowledgeYN != ack) { // if written as a data packet, check to ensure the base station received the message by listening for acknowledgement
    Serial.println("check send");
    checkSend();
  }
}

void checkBattery() {
  int batteryLevel_12bit = analogRead(batteryPin); // record 12-bit ADC value that corresponds to battery voltage divider
  batteryLevel = map(batteryLevel_12bit, 0, 4095, 0, 100); // map value to appropriate value, eventually will be able to calibrate and map to battery level percentage
}

void checkAR(int pin) {
  int arVal = analogRead(pin); // read value from analog sensor
  Serial.print("Moisture Value: ");
  Serial.println(arVal);
  firstByteAR = int2byte(arVal, 1); // save first 4 bits to high byte

  secondByteAR = int2byte(arVal, 2); // save last 8 bits to low byte
  Serial.print("Moisture Value first byte: ");
  Serial.println(firstByteAR);
  Serial.print("Moisture Value second byte: ");
  Serial.println(secondByteAR);
}

void checkLight() {

  if (luxPresent == true) {
    luxRaw = lightSensor.getLux(); // obtain lux value as float from light sensor
    lux = round(luxRaw); // round lux value and save as an int for easier workability
  }

  else {
    luxRaw;
    lux = round(random(40, 100));
  }

  Serial.println("New transmission ----------------------------------------------");
  Serial.print("Raw lux: ");
  Serial.println(luxRaw);
  // light = round(luxRaw*100);
  light = lux;
  firstByteLIGHT = int2byte(light, 1); // save high byte as first byte
  secondByteLIGHT = int2byte(light, 2); // save low byte as second byte
  // Print statements for testing
  Serial.print("Lux: ");
  Serial.println(map(lux, 0, 65535, 0, 16383)); // change this from 0,65535 to 0,4096?
  Serial.print("First byte light: ");
  Serial.println(firstByteLIGHT, HEX);
  Serial.print("Second byte light: ");
  Serial.println(secondByteLIGHT, HEX);
}


void checkSend() { // check to listen for acknowledgement message from basestation for some time
  for (int i = 0; i < 8096; i++) {
    int packetSize = LoRa.parsePacket();
    if (packetSize) {
      Serial.println("packet found in check send");
      while (LoRa.available()) {
        String confirmationMessage = LoRa.readString();
        if (deviceID == confirmationMessage[0]) {
          Serial.println("this message is for me");
          if (confirmationMessage[5] == ack) { // if message contains acknowledgement byte, stop sending
            Serial.print("Confirmation: 0x");
            Serial.println(confirmationMessage[5], BIN);
            confirmed = true; // confirmed = true will stop the field module from continuing transmission
            i = 8096; // when message has been received, exit loop and stop listening for an acknowledgement
          }
        }
      }
    }
  }
}

void checkTemperatureHumidity() {
  uint16_t error;
  char errorMessage[256];

  float temperatureRaw;
  float humidityRaw;

  if (sht4xPresent == true) {

    error = sht4x.measureHighPrecision(temperatureRaw, humidityRaw); // check for either an error or for sensor readings
  }

  if (sht4xPresent == true) {
    if (error) {
      errorToString(error, errorMessage, 256); // if there is an error message, convert to error message
    }
    else { // if no error:
      temperature = round(temperatureRaw * 100); // multiply temperature by 100 and round, to save 2 decimal places but remain an int
      humidity = round(humidityRaw * 100); // multiply humidity by 100 and round, to save 2 decimal places but remain an int


      // Print for checking:
      Serial.print("Temperature:");
      Serial.print(map(temperature, 0, 65535, 0, 16383));
      Serial.print("\t");
      Serial.print("Humidity:");
      Serial.println(map(humidity, 0, 65535, 0, 16383));
    }
  }
  else {
    temperature = round(random(210, 250));
    humidity = round(random(40, 80));
  }
  // Prints for checking:
  Serial.print("Raw Temperature: ");
  Serial.println(temperatureRaw);
  Serial.print("Raw humidity");
  Serial.println(humidityRaw);

  firstByteTEMP = int2byte(temperature, 1); // save high byte as first temperature byte
  secondByteTEMP = int2byte(temperature, 2); // save low byte as second temperature byte

  firstByteHUM = int2byte(humidity, 1); // save high byte as first humidity byte
  secondByteHUM = int2byte(humidity, 2); // save low byte as second humidity byte

  // Print for checking:
  Serial.print("First byte temperature: ");
  Serial.println(firstByteTEMP, HEX);
  Serial.print("Second byte temperature: ");
  Serial.println(secondByteTEMP, HEX);
  Serial.print("First byte humidity: ");
  Serial.println(firstByteHUM, HEX);
  Serial.print("Second byte humidity: ");
  Serial.println(secondByteHUM, HEX);

}

byte int2byte(int a, int byteNum) {
  byte returnByte;
  int b = map(a, 0, 65535, 0, 16383);


  if (byteNum == 1) { // if byte is the high byte
    returnByte = (b >> 7); // bit shift rightward the integer 'a' to save as byte
  }
  else if (byteNum == 2) { // if byte is the low byte
    returnByte = (0x007F & b); // save last 7 bits of int
  }

  return returnByte; // return decomposition as a byte
}
